Package: Boruta
Title: Wrapper Algorithm for All Relevant Feature Selection
Version: 10.0.0
Imports: fru
Suggests: rFerns, randomForest, ranger, survival
Authors@R: c(
  person(
   c("Miron","Bartosz"),"Kursa",
   role=c("aut","cre"),
   email="m@mbq.me",
   comment=c(ORCID="0000-0001-7672-648X")
  ),
  person(
   c("Witold","Remigiusz"),"Rudnicki",
   role="aut"
  )
 )
Description: An all relevant feature selection wrapper algorithm.
 It finds relevant features by comparing original attributes' importance with importance achievable at random, estimated using their permuted copies (shadows).
License: GPL (>= 2)
URL: https://gitlab.com/mbq/Boruta/
RoxygenNote: 7.3.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-05-10 22:18:46 UTC; mbq
Author: Miron Bartosz Kursa [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7672-648X>),
  Witold Remigiusz Rudnicki [aut]
Maintainer: Miron Bartosz Kursa <m@mbq.me>
Repository: CRAN
Date/Publication: 2026-05-16 08:10:16 UTC
Built: R 4.5.2; ; 2026-05-16 12:20:49 UTC; unix

Package: Boruta
Title: Wrapper Algorithm for All Relevant Feature Selection
Version: 10.0.0
Imports: fru
Suggests: rFerns, randomForest, ranger, survival
Authors@R: c(
  person(
   c("Miron","Bartosz"),"Kursa",
   role=c("aut","cre"),
   email="m@mbq.me",
   comment=c(ORCID="0000-0001-7672-648X")
  ),
  person(
   c("Witold","Remigiusz"),"Rudnicki",
   role="aut"
  )
 )
Description: An all relevant feature selection wrapper algorithm.
 It finds relevant features by comparing original attributes' importance with importance achievable at random, estimated using their permuted copies (shadows).
License: GPL (>= 2)
URL: https://gitlab.com/mbq/Boruta/
RoxygenNote: 7.3.3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-05-10 22:18:46 UTC; mbq
Author: Miron Bartosz Kursa [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7672-648X>),
  Witold Remigiusz Rudnicki [aut]
Maintainer: Miron Bartosz Kursa <m@mbq.me>
Repository: CRAN
Date/Publication: 2026-05-16 08:10:16 UTC
Built: R 4.6.0; ; 2026-05-16 09:12:54 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(Boruta,default)
S3method(Boruta,formula)
S3method(plot,Boruta)
S3method(print,Boruta)
export(Boruta)
export(TentativeRoughFix)
export(attStats)
export(conditionalTransdapter)
export(decohereTransdapter)
export(getConfirmedFormula)
export(getImpFerns)
export(getImpFruRaw)
export(getImpFruZ)
export(getImpLegacyRfGini)
export(getImpLegacyRfRaw)
export(getImpLegacyRfZ)
export(getImpRfGini)
export(getImpRfRaw)
export(getImpRfZ)
export(getNonRejectedFormula)
export(getSelectedAttributes)
export(imputeTransdapter)
export(plotImpHistory)

# boruta_py #

[![License](https://img.shields.io/github/license/scikit-learn-contrib/boruta_py)](https://github.com/scikit-learn-contrib/boruta_py/blob/master/LICENSE)
[![PyPI version](https://badge.fury.io/py/Boruta.svg)](https://badge.fury.io/py/Boruta)
[![Anaconda-Server Badge](https://anaconda.org/conda-forge/boruta_py/badges/version.svg)](https://anaconda.org/conda-forge/boruta_py)

This project hosts Python implementations of the [Boruta all-relevant feature selection method](https://www.jstatsoft.org/article/view/v036i11).

[Related blog post](https://danielhomola.com/feature%20selection/phd/borutapy-an-all-relevant-feature-selection-method/)


## How to install ##

Install with `pip`:

```shell
pip install Boruta
```

or with `conda`:

```shell
conda install -c conda-forge boruta_py
```

## Dependencies ##

* numpy
* scipy
* scikit-learn

## How to use ##

Download, import and do as you would with any other scikit-learn method:
* fit(X, y)
* transform(X)
* fit_transform(X, y)

## Description ##

Python implementations of the Boruta R package.

This implementation tries to mimic the scikit-learn interface, so use fit,
transform or fit_transform, to run the feature selection.

For more, see the docs of these functions, and the examples below.

[Original code](https://gitlab.com/mbq/Boruta) and method was authored by Miron B. Kursa.

Boruta is an all relevant feature selection method, while most other are
minimal optimal; this means it tries to find all features carrying
information usable for prediction, rather than finding a possibly compact
subset of features on which some classifier has a minimal error.

Why bother with all relevant feature selection?
When you try to understand the phenomenon that made your data, you should
care about all factors that contribute to it, not just the bluntest signs
of it in context of your methodology (yes, minimal optimal set of features
by definition depends on your classifier choice).


## What's different in BorutaPy? ##

It is the original R package recoded in Python with a few added extra features.
Some improvements include:  

* Faster run times, thanks to scikit-learn

* Scikit-learn like interface

* Compatible with any ensemble method from scikit-learn

* Automatic n_estimator selection

* Ranking of features

* Feature importances are derived from Gini impurity instead of RandomForest R package's MDA. 

For more details, please check the top of the docstring.

We highly recommend using pruned trees with a depth between 3-7.

Also, after playing around a lot with the original code I identified a few areas
where the core algorithm could be improved/altered to make it less strict and
more applicable to biological data, where the Bonferroni correction might be
overly harsh.

__Percentile as threshold__  
The original method uses the maximum of the shadow features as a threshold in
deciding which real feature is doing better than the shadow ones. This could be
overly harsh.

To control this, I added the perc parameter, which sets the
percentile of the shadow features' importances, the algorithm uses as the
threshold. The default of 100 which is equivalent to taking the maximum as the
R version of Boruta does, but it could be relaxed. Note, since this is the
percentile, it changes with the size of the dataset. With several thousands of
features it isn't as stringent as with a few dozens at the end of a Boruta run.


__Two step correction for multiple testing__  
The correction for multiple testing was relaxed by making it a two step
process, rather than a harsh one step Bonferroni correction.

We need to correct firstly because in each iteration we test a number of
features against the null hypothesis (does a feature perform better than
expected by random). For this the Bonferroni correction is used in the original
code which is known to be too stringent in such scenarios (at least for
biological data), and also the original code corrects for n features, even if
we are in the 50th iteration where we only have k<<n features left. For this
reason the first step of correction is the widely used Benjamini Hochberg FDR.

Following that however we also need to account for the fact that we have been
testing the same features over and over again in each iteration with the
same test. For this scenario the Bonferroni is perfect, so it is applied by
dividing the p-value threshold with the current iteration index.

If this two step correction is not required, the two_step parameter has to be
set to False, then (with perc=100) BorutaPy behaves exactly as the R version.

## Parameters ##

__estimator__ : object
   > A supervised learning estimator, with a 'fit' method that returns the
   > feature_importances_ attribute. Important features must correspond to
   > high absolute values in the feature_importances_.

__n_estimators__ : int or string, default = 1000
   > If int sets the number of estimators in the chosen ensemble method.
   > If 'auto' this is determined automatically based on the size of the
   > dataset. The other parameters of the used estimators need to be set
   > with initialisation.

__perc__ : int, default = 100
   > Instead of the max we use the percentile defined by the user, to pick
   > our threshold for comparison between shadow and real features. The max
   > tends to be too stringent. This provides a finer control over this. The
   > lower perc is the more false positives will be picked as relevant but
   > also the less relevant features will be left out. The usual trade-off.
   > The default is essentially the vanilla Boruta corresponding to the max.

__alpha__ : float, default = 0.05
   > Level at which the corrected p-values will get rejected in both correction
   steps.

__two_step__ : Boolean, default = True
  > If you want to use the original implementation of Boruta with Bonferroni
  > correction only set this to False.

__max_iter__ : int, default = 100
   > The number of maximum iterations to perform.

__verbose__ : int, default=0
   > Controls verbosity of output.


## Attributes ##

**n_features_** : int
   > The number of selected features.

**support_** : array of shape [n_features]
   > The mask of selected features - only confirmed ones are True.

**support_weak_** : array of shape [n_features]
  >  The mask of selected tentative features, which haven't gained enough
  >  support during the max_iter number of iterations..

**ranking_** : array of shape [n_features]
  >  The feature ranking, such that ``ranking_[i]`` corresponds to the
  >  ranking position of the i-th feature. Selected (i.e., estimated
  >  best) features are assigned rank 1 and tentative features are assigned
  >  rank 2.


## Examples ##

[![Open example notebook](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/https://raw.githubusercontent.com/scikit-learn-contrib/boruta_py/master/boruta/examples/Madalon_Data_Set.ipynb)
[![Open example notebook](https://b.mineo.app/static/img/open_in_mineo.svg)](https://b.mineo.app/import/https://raw.githubusercontent.com/scikit-learn-contrib/boruta_py/master/boruta/examples/Madalon_Data_Set.ipynb)

```python
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from boruta import BorutaPy

# load X and y
# NOTE BorutaPy accepts numpy arrays only, hence the .values attribute
X = pd.read_csv('examples/test_X.csv', index_col=0).values
y = pd.read_csv('examples/test_y.csv', header=None, index_col=0).values
y = y.ravel()

# define random forest classifier, with utilising all cores and
# sampling in proportion to y labels
rf = RandomForestClassifier(n_jobs=-1, class_weight='balanced', max_depth=5)

# define Boruta feature selection method
feat_selector = BorutaPy(rf, n_estimators='auto', verbose=2, random_state=1)

# find all relevant features - 5 features should be selected
feat_selector.fit(X, y)

# check selected features - first 5 features are selected
feat_selector.support_

# check ranking of features
feat_selector.ranking_

# call transform() on X to filter it down to selected features
X_filtered = feat_selector.transform(X)
```

## References ##

1. Kursa M., Rudnicki W., "Feature Selection with the Boruta Package" Journal of Statistical Software, Vol. 36, Issue 11, Sep 2010

# Supplementary routines for Boruta.

#' Extract attribute statistics
#'
#' \code{attStats} shows a summary of a Boruta run in an attribute-centred way.
#' It produces a data frame containing some importance stats as well as the number of hits that attribute scored and the decision it was given.
#' @param x an object of a class Boruta, from which attribute stats should be extracted.
#' @return A data frame containing, for each attribute that was originally in information system, mean, median, maximal and minimal importance, number of hits normalised to number of importance source runs performed and the decision copied from \code{finalDecision}.
#' @note  When using a Boruta object generated by a \code{\link{TentativeRoughFix}}, the resulting data frame will consist a rough-fixed decision.
#' @note \code{x} has to be made with \code{holdHistory} set to \code{TRUE} for this code to run.
#' @export
attStats<-function(x){
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$ImpHistory))
  stop('Importance history was not stored during the Boruta run.')
 lz<-lapply(1:ncol(x$ImpHistory),function(i) x$ImpHistory[is.finite(x$ImpHistory[,i]),i])
 colnames(x$ImpHistory)->names(lz)
 mr<-lz$shadowMax; lz[1:(length(lz)-3)]->lz
 t(sapply(lz,function(x) c(mean(x),stats::median(x),min(x),max(x),sum(mr[1:length(x)]<x)/length(mr))))->st
 st<-data.frame(st,x$finalDecision)
 names(st)<-c("meanImp","medianImp","minImp","maxImp","normHits","decision")
 return(st)
}

#' Extract names of the selected attributes
#'
#' \code{getSelectedAttributes} returns a vector of names of attributes selected during a Boruta run.
#' @param x an object of a class Boruta, from which relevant attributes names should be extracted.
#' @param withTentative if set to \code{TRUE}, Tentative attributes will be also returned.
#' @return A character vector with names of the relevant attributes.
#' @export
getSelectedAttributes<-function(x,withTentative=FALSE){
 stopifnot(inherits(x,'Boruta'))
 names(x$finalDecision)[
  x$finalDecision%in%(if(!withTentative) "Confirmed" else c("Confirmed","Tentative"))
 ]
}

#' Rough fix of Tentative attributes
#'
#' In some circumstances (too short Boruta run, unfortunate mixing of shadow attributes, tricky dataset\ldots), Boruta can leave some attributes Tentative.
#' \code{TentativeRoughFix} performs a simplified, weaker test for judging such attributes.
#' @param x an object of a class Boruta.
#' @param averageOver Either number of last importance source runs to
#' average over or Inf for averaging over the whole Boruta run.
#' @return A Boruta class object with modified \code{finalDecision} element.
#' Such object has few additional elements:
#' \item{originalDecision}{Original \code{finalDecision}.}
#' \item{averageOver}{Copy of \code{averageOver} parameter.}
#' @details Function claims as Confirmed those attributes that
#' have median importance higher than the median importance of
#' maximal shadow attribute, and the rest as Rejected.
#' Depending of the user choice, medians for the test
#' are count over last round, all rounds or N last
#' importance source runs.
#' @note This function should be used only when strict decision is
#' highly desired, because this test is much weaker than Boruta
#' and can lower the confidence of the final result.
#' @note \code{x} has to be made with \code{holdHistory} set to
#' \code{TRUE} for this code to run.
#' @export
TentativeRoughFix<-function(x,averageOver=Inf){
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$ImpHistory))
  stop('Importance history was not stored during the Boruta run.')
 if(!is.numeric(averageOver))
  stop('averageOver should be a numeric vector.')
 if(length(averageOver)!=1)
  stop('averageOver should be a one-element vector.')
 if(averageOver<1)
  stop('averageOver should be positive.')

 tentIdx<-which(x$finalDecision=='Tentative')
 if(length(tentIdx)==0){
  warning('There are no Tentative attributes! Returning original object.')
  return(x)
 }

 nRuns<-dim(x$ImpHistory)[1]

 if(averageOver>nRuns)
  averageOver<-nRuns

 impHistorySubset<-x$ImpHistory[(nRuns-averageOver+1):nRuns,]
 medianTentImp<-apply(impHistorySubset[,tentIdx,drop=FALSE],2,stats::median)
 medianShaMaxImp<-stats::median(impHistorySubset[,'shadowMax'])
 medianTentImp>medianShaMaxImp->toOrdain

 ans<-x
 ans$roughfixed<-TRUE
 ans$averageOver<-averageOver
 ans$originalDecision<-x$finalDecision
 ans$finalDecision[tentIdx[toOrdain]]<-'Confirmed'
 ans$finalDecision[tentIdx[!toOrdain]]<-'Rejected'

 return(ans)
}

##generateCol is internally used by plot.Boruta and plotImpHistory
generateCol<-function(x,colCode,col,numShadow){
 #Checking arguments
 if(is.null(col) & length(colCode)!=4)
  stop('colCode should have 4 elements.')
 #Generating col
 if(is.null(col)){
  rep(colCode[4],length(x$finalDecision)+numShadow)->cc
  cc[c(x$finalDecision=='Confirmed',rep(FALSE,numShadow))]<-colCode[1]
  cc[c(x$finalDecision=='Tentative',rep(FALSE,numShadow))]<-colCode[2]
  cc[c(x$finalDecision=='Rejected',rep(FALSE,numShadow))]<-colCode[3]
  col=cc
 }
 return(col)
}

#' Plot Boruta object
#'
#' Default plot method for Boruta objects, showing boxplots of attribute importances over run.
#' @method plot Boruta
#' @param x an object of a class Boruta.
#' @param colCode a vector containing colour codes for attribute decisions, respectively Confirmed, Tentative, Rejected and shadow.
#' @param sort controls whether boxplots should be ordered, or left in original order.
#' @param whichShadow a logical vector controlling which shadows should be drawn; switches respectively max shadow, mean shadow and min shadow.
#' @param col standard \code{col} attribute. If given, suppresses effects of \code{colCode}.
#' @param xlab X axis label that will be passed to \code{\link{boxplot}}.
#' @param ylab Y axis label that will be passed to \code{\link{boxplot}}.
#' @param ... additional graphical parameter that will be passed to \code{\link{boxplot}}.
#' @note If \code{col} is given and \code{sort} is \code{TRUE}, the \code{col} will be permuted, so that its order corresponds to attribute order in \code{ImpHistory}.
#' @note This function will throw an error when \code{x} lacks importance history, i.e., was made with \code{holdHistory} set to \code{FALSE}.
#' @return Invisible copy of \code{x}.
#' @export
plot.Boruta<-function(x,colCode=c('green','yellow','red','blue'),sort=TRUE,whichShadow=c(TRUE,TRUE,TRUE),
  col=NULL,xlab='Attributes',ylab='Importance',...){
 #Checking arguments
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$ImpHistory))
  stop('Importance history was not stored during the Boruta run.')

 #Removal of -Infs and conversion to a list
 lz<-lapply(1:ncol(x$ImpHistory),function(i) x$ImpHistory[is.finite(x$ImpHistory[,i]),i])
 colnames(x$ImpHistory)->names(lz)

 #Selection of shadow meta-attributes
 numShadow<-sum(whichShadow)
 lz[c(rep(TRUE,length(x$finalDecision)),whichShadow)]->lz

 #Generating color vector
 col<-generateCol(x,colCode,col,numShadow)

 #Ordering boxes due to attribute median importance
 if(sort){
  ii<-order(sapply(lz,stats::median))
  lz[ii]->lz; col<-col[ii]
 }

 #Final plotting
 graphics::boxplot(lz,xlab=xlab,ylab=ylab,col=col,...)
 invisible(x)
}

#' Plot Boruta object as importance history
#'
#' Alternative plot method for Boruta objects, showing matplot of attribute importances over run.
#' @param x an object of a class Boruta.
#' @param colCode a vector containing colour codes for attribute decisions, respectively Confirmed, Tentative, Rejected and shadow.
#' @param col standard \code{col} attribute, passed to \code{\link{matplot}}. If given, suppresses effects of \code{colCode}.
#' @param type Plot type that will be passed to \code{\link{matplot}}.
#' @param lty Line type that will be passed to \code{\link{matplot}}.
#' @param pch Point mark type that will be passed to \code{\link{matplot}}.
#' @param xlab X axis label that will be passed to \code{\link{matplot}}.
#' @param ylab Y axis label that will be passed to \code{\link{matplot}}.
#' @param ... additional graphical parameter that will be passed to \code{\link{matplot}}.
#' @note This function will throw an error when \code{x} lacks importance history, i.e., was made with \code{holdHistory} set to \code{FALSE}.
#' @return Invisible copy of \code{x}.
#' @export
plotImpHistory<-function(x,colCode=c('green','yellow','red','blue'),col=NULL,type="l",lty=1,pch=0,
  xlab='Classifier run',ylab='Importance',...){
 #Checking arguments
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$ImpHistory))
  stop('Importance history was not stored during the Boruta run.')
 col<-generateCol(x,colCode,col,3)

 #Final plotting
 graphics::matplot(0:(nrow(x$ImpHistory)-1),x$ImpHistory,xlab=xlab,ylab=ylab,col=col,type=type,lty=lty,pch=pch,...)
 invisible(x)
}

#' Export Boruta result as a formula
#'
#' Functions which convert the Boruta selection into a formula, so that it could be passed further to other functions.
#' @param x an object of a class Boruta, made using a formula interface.
#' @return Formula, corresponding to the Boruta results.
#' \code{getConfirmedFormula} returns only Confirmed attributes, \code{getNonRejectedFormula} also adds Tentative ones.
#' @note This operation is possible only when Boruta selection was invoked using a formula interface.
#' @rdname getFormulae
#' @export
getConfirmedFormula<-function(x){
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$call[["formula"]]))
  stop('The model for this Boruta run was not a formula.')
 deparse(x$call[["formula"]][[2]])->dec
 preds<-paste(names(x$finalDecision)[x$finalDecision=='Confirmed'],collapse="+")
 return(stats::as.formula(sprintf('%s~%s',dec,preds)))
}

#' @rdname getFormulae
#' @export
getNonRejectedFormula<-function(x){
 stopifnot(inherits(x,'Boruta'))
 if(is.null(x$call[["formula"]]))
  stop('The model for this Boruta run was not a formula.')
 deparse(x$call[["formula"]][[2]])->dec
 preds<-paste(names(x$finalDecision)[x$finalDecision!='Rejected'],collapse="+")
 return(stats::as.formula(sprintf('%s~%s',dec,preds)))
}

make_df<-function(formula,data,enc){
 if(missing(enc)) enc<-parent.frame()

 if(missing(data)){
  env<-enc
  have<-c()
 }else{
  if(!is.data.frame(data)) stop("data must be a data.frame")
  env<-data
  have<-names(data)
 }
 to_delete<-c()
 to_add<-c()
 new_features<-list()
 has_dot<-FALSE

 f<-stats::as.formula(formula)

 Ye<-f[[2]]
 Y<-eval(Ye,env,enc)
 Yn<-deparse(Ye)
 if(is.symbol(Ye)&&(Yn%in%have)) to_delete<-Yn

 f[[3]]->f
 while(TRUE){
  if(length(f)==3){
   oper<-deparse(f[[1]])
   element<-f[[3]]
   if(length(element)!=1)
    if(!identical(element[[1]],quote(I)))
     stop("Invalid sub-expression ",deparse(element))
  }else{
   #This is the last element
   oper<-'+'
   element<-f
  }

  if(oper=='-'){
   if(!is.symbol(element))
    stop(sprintf("Cannot omit something that is not a feature name (%s)",deparse(element)))
   element<-deparse(element)
   if(!(element%in%have))
    stop(sprintf("Cannot omit %s which is not in data",element))
   to_delete<-c(to_delete,element)
  }else if(oper=='+'){
   if(is.symbol(element)){
    deparse(element)->en
    if(en=='.'){
     if(missing(data)) stop("Cannot use `.` without data")
     has_dot<-TRUE
    }else{
     if(en%in%have){
      to_add<-c(to_add,en)
     }else{
      eval(element,env,enc)->val
      new_features<-c(new_features,stats::setNames(list(val),en))
     }
    }
   }else{
    eval(element,env,enc)->val
    new_features<-c(new_features,stats::setNames(list(val),deparse(element)))
   }
  }else
   stop(sprintf("Invalid operator `%s`; only `+` & `-` allowed",oper))
  if(length(f)<3) break
  f<-f[[2]]
 }

 if(has_dot){
  to_delete<-setdiff(to_delete,to_add)
  X<-data
  if(length(to_delete)>0)
   X<-X[,setdiff(names(X),to_delete),drop=FALSE]
  if(length(new_features)>0)
   X<-data.frame(X,new_features)
 }else{
  if(!missing(data)){
   X<-data[,setdiff(to_add,to_delete),drop=FALSE]
   if(length(new_features)>0)
    X<-data.frame(X,new_features)
  }else if(length(new_features)>0){
    X<-data.frame(new_features)
  } #else can't happen because of formula syntax properties; Y~ is invalid
 }
 list(X=X,Y=Y)
}

# Importance sources

#' randomForest importance adapters
#'
#' Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
#' \code{getImpLegacyRfZ} generates default, normalized permutation importance, \code{getImpLegacyRfRaw} raw permutation importance, finally \code{getImpLegacyRfGini} generates Gini index importance, all using \code{\link[randomForest]{randomForest}} as a Random Forest algorithm implementation.
#' @name getImpLegacyRf
#' @rdname getImpLegacyRf
#' @aliases getImpLegacyRfZ getImpLegacyRfGini getLegacyImpRfRaw
#' @note The \code{getImpLegacyRfZ} function was a default importance source in Boruta versions prior to 5.0; since then \code{\link[ranger]{ranger}} Random Forest implementation is used instead of \code{\link[randomForest]{randomForest}}, for speed, memory conservation and an ability to use multithreading.
#' Both importance sources should generally lead to the same results, yet there are differences.
#'
#' Most notably, ranger by default treats factor attributes as ordered (and works very slow if instructed otherwise with \code{respect.unordered.factors=TRUE}); on the other hand it lifts 32 levels limit specific to \code{\link[randomForest]{randomForest}}.
#' To this end, Boruta decision for factor attributes may be different.
#'
#' Random Forest methods has two main parameters, number of attributes tried at each split and the number of trees in the forest; first one is called \code{mtry} in both implementations, but the second \code{ntree} in \code{\link[randomForest]{randomForest}} and \code{num.trees} in \code{\link[ranger]{ranger}}.
#' To this end, to maintain compatibility, \code{getImpRf*} functions still accept \code{ntree} parameter relaying it into \code{num.trees}.
#' Still, both parameters take the same defaults in both implementations (square root of the number all all attributes and 500 respectively).
#'
#' Moreover, \code{\link[ranger]{ranger}} brings some addition capabilities to Boruta, like analysis of survival problems or sticky variables which are always considered on splits.
#'
#' Finally, the results for the same PRNG seed will be different.
#' @param x data frame of predictors including shadows.
#' @param y response vector.
#' @param ... parameters passed to the underlying \code{\link[randomForest]{randomForest}} call; they are relayed from \code{...} of \code{\link{Boruta}}.
#' @export
getImpLegacyRfZ<-function(x,y,...){
 randomForest::randomForest(x,y,
  importance=TRUE,keep.forest=FALSE,...)->rf
 randomForest::importance(rf,1,scale=TRUE)[,1]
}
comment(getImpLegacyRfZ)<-'randomForest normalized permutation importance'

#' @rdname getImpLegacyRf
#' @export
getImpLegacyRfRaw<-function(x,y,...){
 randomForest::randomForest(x,y,
  importance=TRUE,keep.forest=FALSE,...)->rf
 randomForest::importance(rf,1,scale=FALSE)[,1]
}
comment(getImpLegacyRfRaw)<-'randomForest raw permutation importance'

#' @rdname getImpLegacyRf
#' @export
getImpLegacyRfGini<-function(x,y,...){
 randomForest::randomForest(x,y,
  keep.forest=FALSE,...)->rf
 randomForest::importance(rf,2,scale=FALSE)[,1]
}
comment(getImpLegacyRfGini)<-'randomForest Gini index importance'

#' Fru Random Forest importance adapters
#'
#' Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
#' \code{getImpFruZ} generates default, normalized permutation importance while \code{getImpFruRaw} raw permutation importance.
#' Fru does not support Gini index importance.
#' @name getImpFru
#' @rdname getImpFru
#' @aliases getImpFruZ getImFruRaw
#' @param x data frame of predictors including shadows.
#' @param y response vector.
#' @param ntree  Number of trees in the forest; copied into \code{\link[fru]{fru}}'s native \code{trees}, put to retain transparent compatibility with randomForest.
#' @param trees  Number of trees in the forest, as according to \code{\link[fru]{fru}}'s nomenclature. If not given, set to \code{ntree} value. If both are given, \code{trees} takes precedence.
#' @param num.threads Number of computing threads to use; copied into \code{\link[fru]{fru}}'s native \code{threads}, put to retain transparent compatibility with ranger.
#' @param threads Number of computing threads to use, as according to \code{\link[fru]{fru}}'s nomenclature.
#' If not given, set to \code{num.threads}.
#' If both are given, \code{threads} takes precedence.
#' Default value of zero means all available threads.
#' @param ... parameters passed to the underlying \code{\link[fru]{fru}} call; they are relayed from \code{...} of \code{\link{Boruta}}.
#' @note In prior versions of Boruta, other implementations of Random Forest were used as a default; prior to 5.0.0, randomForest package was used, and the default adapter used back then is available as \code{\link{getImpLegacyRf}}.
#' Prior to 10.0, ranger package was used with \code{\link{getImpRfZ}}.
#' @export
getImpFruZ<-function(x,y,ntree=500,trees=ntree,num.threads=0,threads=num.threads,...){
 if(inherits(y,"Surv"))
  stop("Fru doesn't support censored data; try getImpRfZ as an importance source")
 fru::importance(fru::fru(x,y,trees=trees,importance=TRUE,threads=threads,...),scale=TRUE)
}
comment(getImpFruZ)<-'fru normalized permutation importance'

#' @rdname getImpFru
#' @export
getImpFruRaw<-function(x,y,ntree=500,trees=ntree,num.threads=0,threads=num.threads,...){
 if(inherits(y,"Surv"))
  stop("Fru doesn't support censored data; try getImpRfRaw as an importance source")
 fru::importance(fru::fru(x,y,trees=trees,importance=TRUE,threads=threads,...),scale=FALSE)
}
comment(getImpFruZ)<-'fru raw permutation importance'

#' Ranger Random Forest importance adapters
#'
#' Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
#' \code{getImpRfZ} generates default, normalized permutation importance, \code{getImpRfRaw} raw permutation importance, finally \code{getImpRfGini} generates Gini index importance.
#' @name getImpRf
#' @rdname getImpRf
#' @aliases getImpRfZ getImpRfGini getImpRfRaw
#' @param x data frame of predictors including shadows.
#' @param y response vector.
#' @param ntree  Number of trees in the forest; copied into \code{\link[ranger]{ranger}}'s native num.trees, put to retain transparent compatibility with randomForest.
#' @param num.trees  Number of trees in the forest, as according to \code{\link[ranger]{ranger}}'s nomenclature. If not given, set to \code{ntree} value. If both are given, \code{num.trees} takes precedence.
#' @param ... parameters passed to the underlying \code{\link[ranger]{ranger}} call; they are relayed from \code{...} of \code{\link{Boruta}}.
#' @note \code{getImpRfZ} was the default importance source between versions 5.0.0 and 10.0.0. Prior to Boruta 5.0, \code{getImpLegacyRfZ} function was a default; see \link{getImpLegacyRf} for more details.
#' @export
getImpRfZ<-function(x,y,ntree=500,num.trees=ntree,...){
 if(inherits(y,"Surv")){
  return(ranger::ranger(x=x,y=y,
   num.trees=num.trees,importance="permutation",
   scale.permutation.importance=TRUE,
   write.forest=FALSE,...)$variable.importance)
 }
 ranger::ranger(x=x,y=y,
  num.trees=num.trees,importance="permutation",
  scale.permutation.importance=TRUE,
  write.forest=FALSE,...)$variable.importance
}
comment(getImpRfZ)<-'ranger normalized permutation importance'

#' @rdname getImpRf
#' @export
getImpRfGini<-function(x,y,ntree=500,num.trees=ntree,...){
 if(inherits(y,"Surv"))
  stop("Ranger cannot produce Gini importance for survival problems.")
 ranger::ranger(x=x,y=y,
  num.trees=num.trees,importance="impurity",
  scale.permutation.importance=FALSE,
  write.forest=FALSE,...)$variable.importance
}
comment(getImpRfGini)<-'ranger Gini index importance'

#' @rdname getImpRf
#' @export
getImpRfRaw<-function(x,y,ntree=500,num.trees=ntree,...){
 if(inherits(y,"Surv")){
  return(ranger::ranger(x=x,y=y,
   num.trees=num.trees,importance="permutation",
   write.forest=FALSE,...)$variable.importance)
 }
 ranger::ranger(x=x,y=y,
  num.trees=num.trees,importance="permutation",
  scale.permutation.importance=FALSE,
  write.forest=FALSE,...)$variable.importance
}
comment(getImpRfRaw)<-'ranger raw permutation importance'

#' Random Ferns importance
#'
#' This function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
#' @param x data frame of predictors including shadows.
#' @param y response vector.
#' @param ... parameters passed to the underlying \code{\link[rFerns]{rFerns}} call; they are relayed from \code{...} of \code{\link{Boruta}}.
#' @export
#' @note Random Ferns importance calculation should be much faster than using Random Forest; however, one must first optimize the value of the \code{depth} parameter and
#' it is quite likely that the number of ferns in the ensemble required for the importance to converge will be higher than the number of trees in case of Random Forest.
getImpFerns<-function(x,y,...){
 f<-rFerns::rFerns(x,y,
  saveForest=FALSE,importance=TRUE,...)
 f$importance[,1]
}
comment(getImpFerns)<-'rFerns importance'

# Core of Boruta

#' @export
#' @rdname Boruta
Boruta<-function(x,...)
 UseMethod("Boruta")

#' Feature selection with the Boruta algorithm
#'
#' Boruta is an all relevant feature selection wrapper algorithm, capable of working with any classification method that output variable importance measure (VIM); by default, Boruta uses Random Forest.
#' The method performs a top-down search for relevant features by comparing original attributes' importance with importance achievable at random, estimated using their permuted copies, and progressively eliminating irrelevant features to stabilise that test.
#' @rdname Boruta
#' @method Boruta default
#' @param x data frame of predictors.
#' @param y response vector; factor for classification, numeric vector for regression, \code{Surv} object for survival (support depends on importance adapter capabilities).
#' @param getImp function used to obtain attribute importance, called importance adapter in this documentation.
#' The default is getImpFruZ, which runs Random Forest implementation from the \code{fru} package and gathers Z-scores of mean decrease accuracy measure.
#' It should return a numeric vector of a size identical to the number of columns of its first argument, containing importance measure of respective attributes.
#' Any order-preserving transformation of this measure will yield the same result.
#' It is assumed that more important attributes get higher importance. +-Inf are accepted, NaNs and NAs are treated as 0s, with a warning.
#' @param pValue confidence level. Default value should be used.
#' @param mcAdj if set to \code{TRUE}, a multiple comparisons adjustment using the Bonferroni method will be applied. Default value should be used; older (1.x and 2.x) versions of Boruta were effectively using \code{FALSE}.
#' @param maxRuns maximal number of importance source runs.
#' You may increase it to resolve attributes left Tentative.
#' @param holdHistory if set to \code{TRUE}, the full history of importance is stored and returned as the \code{ImpHistory} element of the result.
#' Can be used to decrease a memory footprint of Boruta in case this side data is not used, especially when the number of attributes is huge; yet it disables plotting of such made \code{Boruta} objects and the use of the \code{\link{TentativeRoughFix}} function.
#' @param doTrace verbosity level. 0 means no tracing, 1 means reporting decision about each attribute as soon as it is justified, 2 means the same as 1, plus reporting each importance source run, 3 means the same as 2, plus reporting of hits assigned to yet undecided attributes.
#' @param ... additional parameters passed to \code{getImp}.
#' @return An object of class \code{Boruta}, which is a list with the following components:
#' \item{finalDecision}{a factor of three value: \code{Confirmed}, \code{Rejected} or \code{Tentative}, containing final result of feature selection.}
#' \item{ImpHistory}{a data frame of importances of attributes gathered in each importance source run.
#' Beside predictors' importances, it contains maximal, mean and minimal importance of shadow attributes in each run.
#' Rejected attributes get \code{-Inf} importance.
#' Set to \code{NULL} if \code{holdHistory} was given \code{FALSE}.}
#' \item{timeTaken}{time taken by the computation.}
#' \item{impSource}{string describing the source of importance, equal to a comment attribute of the \code{getImp} argument.}
#' \item{call}{the original call of the \code{Boruta} function.}
#' @details Boruta iteratively compares importances of attributes with importances of shadow attributes, created by shuffling original ones.
#' Attributes that have significantly worst importance than shadow ones are being consecutively dropped.
#' On the other hand, attributes that are significantly better than shadows are admitted to be Confirmed.
#' Shadows are re-created in each iteration.
#' Algorithm stops when only Confirmed attributes are left, or when it reaches \code{maxRuns} importance source runs.
#' If the second scenario occurs, some attributes may be left without a decision.
#' They are claimed Tentative.
#' You may try to extend \code{maxRuns} or lower \code{pValue} to clarify them, but in some cases their importances do fluctuate too much for Boruta to converge.
#' Instead, you can use \code{\link{TentativeRoughFix}} function, which will perform other, weaker test to make a final decision, or simply treat them as undecided in further analysis.
#' @references Miron B. Kursa, Witold R. Rudnicki (2010). Feature Selection with the Boruta Package.
#' \emph{Journal of Statistical Software, 36(11)}, p. 1-13.
#' URL: \doi{10.18637/jss.v036.i11}
#' @export
Boruta.default<-function(x,y,pValue=0.01,mcAdj=TRUE,maxRuns=100,doTrace=0,holdHistory=TRUE,getImp=getImpFruZ,...){
 #Timer starts... now!
 timeStart<-Sys.time()

 #Extract the call to store in output
 cl<-match.call()
 cl[[1]]<-as.name('Boruta')

 #Convert x into a data.frame
 if(!is.data.frame(x))
  x<-data.frame(x)

 ##Some checks on x & y
 if(length(grep('^shadow',names(x)))>0)
  stop('Attributes with names starting from "shadow" are reserved for internal use. Please rename them.')
 if(maxRuns<11)
  stop('maxRuns must be greater than 10.')

 ##Expands the information system with newly built random attributes and calculates importance
 addShadowsAndGetImp<-function(decReg,runs){
  #xSha is going to be a data frame with shadow attributes; time to init it.
  xSha<-x[,decReg!="Rejected",drop=F]
  while(dim(xSha)[2]<5) xSha<-cbind(xSha,xSha); #There must be at least 5 random attributes.

  #Now, we permute values in each attribute
  nSha<-ncol(xSha)
  data.frame(lapply(xSha,sample))->xSha
  names(xSha)<-paste('shadow',1:nSha,sep="")

  #Notifying user of our progress
  if(doTrace>1)
   message(sprintf(' %s. run of importance source...',runs))

  #Calling importance source; "..." can be used by the user to pass rf attributes (for instance ntree)
  impRaw<-getImp(cbind(x[,decReg!="Rejected"],xSha),y,...)
  if(!is.numeric(impRaw))
   stop("getImp result is not a numeric vector. Please check the given getImp function.")
  if(length(impRaw)!=sum(decReg!="Rejected")+ncol(xSha))
   stop("getImp result has a wrong length. Please check the given getImp function.")
  if(any(is.na(impRaw)|is.nan(impRaw))){
   impRaw[is.na(impRaw)|is.nan(impRaw)]<-0
   warning("getImp result contains NA(s) or NaN(s); replacing with 0(s), yet this is suspicious.")
  }

  #Importance must have Rejected attributes put on place and filled with -Infs
  imp<-rep(-Inf,nAtt+nSha);names(imp)<-c(attNames,names(xSha))
  impRaw->imp[c(decReg!="Rejected",rep(TRUE,nSha))]
  shaImp<-imp[(nAtt+1):length(imp)];imp[1:nAtt]->imp

  return(list(imp=imp,shaImp=shaImp))
 }

 ##Assigns hits
 assignHits<-function(hitReg,curImp){
  curImp$imp>max(curImp$shaImp)->hits
  if(doTrace>2){
   uncMask<-decReg=="Tentative"
   intHits<-sum(hits[uncMask])
   if(intHits>0)
    message(sprintf("Assigned hit to %s attribute%s out of %s undecided.",sum(hits[uncMask]),if(intHits==1) "" else "s",sum(uncMask)))
   else
    message("None of undecided attributes scored a hit.")
  }
  hitReg[hits]<-hitReg[hits]+1
  return(hitReg)
 }

 ##Checks whether number of hits is significant
 doTests<-function(decReg,hitReg,runs){
  pAdjMethod<-ifelse(mcAdj[1],'bonferroni','none')
  #If attribute is significantly more frequent better than shadowMax, its claimed Confirmed
  toAccept<-stats::p.adjust(stats::pbinom(hitReg-1,runs,0.5,lower.tail=FALSE),method=pAdjMethod)<pValue
  (decReg=="Tentative" & toAccept)->toAccept

  #If attribute is significantly more frequent worse than shadowMax, its claimed Rejected (=irrelevant)
  toReject<-stats::p.adjust(stats::pbinom(hitReg,runs,0.5,lower.tail=TRUE),method=pAdjMethod)<pValue
  (decReg=="Tentative" & toReject)->toReject

  #Update decReg
  decReg[toAccept]<-"Confirmed";"Rejected"->decReg[toReject]

  #Report progress
  if(doTrace>0){
   nAcc<-sum(toAccept)
   nRej<-sum(toReject)
   nLeft<-sum(decReg=="Tentative")
   if(nAcc+nRej>0)
    message(sprintf("After %s iterations, +%s: ",runs,format(difftime(Sys.time(),timeStart),digits=2)))
   if(nAcc>0)
    message(sprintf(" confirmed %s attribute%s: %s",
     nAcc,ifelse(nAcc==1,'','s'),.attListPrettyPrint(attNames[toAccept])))
   if(nRej>0)
    message(sprintf(" rejected %s attribute%s: %s",
     nRej,ifelse(nRej==1,'','s'),.attListPrettyPrint(attNames[toReject])))
   if(nAcc+nRej>0)
    if(nLeft>0){
     message(sprintf(" still have %s attribute%s left.\n",
      nLeft,ifelse(nLeft==1,'','s')))
    }else{
     if(nAcc+nRej>0) message(" no more attributes left.\n")
    }
  }
  return(decReg)
 }

 ##Creating some useful constants
 nAtt<-ncol(x); nrow(x)->nObjects
 attNames<-names(x); c("Tentative","Confirmed","Rejected")->confLevels

 ##Initiate state
 decReg<-factor(rep("Tentative",nAtt),levels=confLevels)
 hitReg<-rep(0,nAtt);names(hitReg)<-attNames
 impHistory<-list()
 runs<-0

 ##Main loop

 while(any(decReg=="Tentative") && (runs+1->runs)<maxRuns){
  curImp<-addShadowsAndGetImp(decReg,runs)
  hitReg<-assignHits(hitReg,curImp)
  decReg<-doTests(decReg,hitReg,runs)

  #If needed, update impHistory with scores obtained in this iteration
  if(holdHistory){
   imp<-c(curImp$imp,
    shadowMax=max(curImp$shaImp),
    shadowMean=mean(curImp$shaImp),
    shadowMin=min(curImp$shaImp))
   impHistory<-c(impHistory,list(imp))
  }
 }

 ##Building result
 impHistory<-do.call(rbind,impHistory)
 names(decReg)<-attNames
 ans<-list(finalDecision=decReg,ImpHistory=impHistory,
   pValue=pValue,maxRuns=maxRuns,light=TRUE,mcAdj=mcAdj,
   timeTaken=Sys.time()-timeStart,roughfixed=FALSE,call=cl,
   impSource=comment(getImp))

 "Boruta"->class(ans)
 return(ans)
}

.attListPrettyPrint<-function(x,limit=5){
 x<-sort(x)
 if(length(x)<limit+1)
  return(sprintf("%s;",paste(x,collapse=", ")))
 sprintf("%s and %s more;",paste(utils::head(x,limit),collapse=", "),length(x)-limit)
}

#' @rdname Boruta
#' @method Boruta formula
#' @param formula alternatively, formula describing model to be analysed.
#' @param data in which to interpret formula.
#' @export
Boruta.formula<-function(formula,data,...){
 make_df(formula,data,parent.frame())->d

 ##Run Boruta
 ans<-Boruta.default(d$X,d$Y,...)
 ans$call<-match.call()
 ans$call[[1]]<-as.name('Boruta')
 formula->ans$call[["formula"]]
 return(ans)
}

#' Print Boruta object
#'
#' Print method for the Boruta objects.
#' @method print Boruta
#' @param x an object of a class Boruta.
#' @param ... additional arguments passed to \code{\link{print}}.
#' @return Invisible copy of \code{x}.
#' @export
print.Boruta<-function(x,...){
 stopifnot(inherits(x,'Boruta'))
 cat(paste('Boruta performed ',dim(x$ImpHistory)[1],' iterations in ',format(x$timeTaken),'.\n',sep=''))
 if(x$roughfixed) cat(paste('Tentatives roughfixed over the last ',x$averageOver,' iterations.\n',sep=''))
 if(sum(x$finalDecision=='Confirmed')==0){
  cat(' No attributes deemed important.\n')} else {
  writeLines(strwrap(paste(sum(x$finalDecision=='Confirmed'),' attributes confirmed important: ',
   .attListPrettyPrint(names(x$finalDecision[x$finalDecision=='Confirmed']))),indent=1))
 }
 if(sum(x$finalDecision=='Rejected')==0){
  cat(' No attributes deemed unimportant.\n')} else {
  writeLines(strwrap(paste(sum(x$finalDecision=='Rejected'),' attributes confirmed unimportant: ',
   .attListPrettyPrint(names(x$finalDecision[x$finalDecision=='Rejected']))),indent=1))
 }
 if(sum(x$finalDecision=='Tentative')!=0){
  writeLines(strwrap(paste(sum(x$finalDecision=='Tentative'),' tentative attributes left: ',
   .attListPrettyPrint(names(x$finalDecision[x$finalDecision=='Tentative']))),indent=1))
 }
 invisible(x)
}

#' Small redundant XOR data
#'
#' A synthetic data set with 32 rows corresponding to all combinations of values of five logical features, A, B, N1, N2 and N3.
#' The decision Y is equal to A xor B, hence N1--N3 are irrelevant attributes.
#' The set also contains 3 additional features, A or B (AoB), A and B (AnB) and not A (nA), which provide a redundant, but still relevant way to reconstruct Y.
#'
#' This is set is an easy way to demonstrate the difference between all relevant feature selection methods, which should select all features except N1--N3, and minimal optimal ones, which will probably ignore most of them.
#' @format A data frame with 8 predictors, 4 relevant: A, B, AoB, AnB and nA, as well as 3 irrelevant N1, N2 and N3, and decision attribute Y.
#' @source \url{https://blog.mbq.me/relevance-and-redundancy/}
#' @usage data(srx)
"srx"

# Input transformations which compose with importance adapters

fixna<-function(x){
 nm<-is.na(x)
 if(!any(nm)) return(x)
 if(all(nm)) return(rep(0,length(x)))
 x[nm]<-sample(x[!nm],sum(nm),replace=TRUE)
 x
}

#' Impute transdapter
#'
#' Wraps the importance adapter to accept NAs in input.
#' 
#' @param adapter importance adapter to transform.
#' @return transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
#' @note An all-NA feature will be converted to all zeroes, which should be ok as a totally non-informative value with most methods, but it is not universally correct.
#' Ideally, one should avoid having such features in input altogether.
#' @export
imputeTransdapter<-function(adapter=getImpFruZ){
 composition<-function(x,y,...)
  adapter(
   data.frame(lapply(x,fixna)),
   fixna(y),
   ...
  )
 comment(composition)<-sprintf("%s, wrapped into imputation transdapter",comment(adapter))
 composition
}
 
#' Decohere transdapter
#'
#' Applies the decoherence transformation to the input, destroying all multivariate interactions.
#' It will trash the Boruta result, only apply if you know what are you doing!
#' Works only for categorical decision.
#' 
#' @param adapter importance adapter to transform.
#' @return transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
#' @export
decohereTransdapter<-function(adapter=getImpFruZ){
 composition<-function(x,y,...){
  stopifnot(is.factor(y))
  mix<-function(x) as.data.frame(lapply(x,sample),row.names=rownames(x))
  unsplit(lapply(split(x,y),mix),y)->xd
  adapter(
   xd,
   y,
   ...
  )
 }
 comment(composition)<-sprintf("%s, wrapped into decoherence transdapter",comment(adapter))
 composition
}

#' Conditional transdapter
#'
#' Applies downstream importance source on a given object strata and averages their outputs.
#' 
#' @param groups groups.
#' @param adapter importance adapter to transform.
#' @return transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
#' @export
conditionalTransdapter<-function(groups,adapter=getImpFruZ){
 as.numeric(table(groups))/length(groups)->w
 stopifnot(is.factor(groups))
 composition<-function(x,y,...)
  colSums(w*t(sapply(levels(groups),function(l) adapter(x[groups==l,],y[groups==l],...))))
 comment(composition)<-sprintf("%s, wrapped into conditional transdapter",comment(adapter))
 composition
}
 

\documentclass{article}
%\VignetteIndexEntry{Importance transdapters}
\usepackage[utf8]{inputenc}

\title{Importance transdapters}
\author{Miron B. Kursa}

\begin{document}
\bibliographystyle{plain}
\maketitle

\section{Overview}
Boruta is essentially a significance test for the importance of features generated by some importance source.
This source is usually the normalised permutational importance from a Random Forest \cite{Breiman2001} model, although other options are possible.
The interface between Boruta and importance source is called \emph{importance adapter}, which is an R function fed into \texttt{getImp} argument of the \texttt{Boruta} function.
They should expect three arguments, \texttt{x}, for predictors table with shadows, \texttt{y}, for the decision, and \texttt{...}, for arbitrary additional arguments relayed from the Boruta call.
For instance, \texttt{Boruta(Species~.,iris,num.threads=1)} is going to pass the \texttt{num.threads} argument to the adapter function, which in turn will append it to its call to \texttt{ranger}.
The adapter output is expected to be a numerical vector of the same length as \texttt{ncol(x)}, with $i$-th element being the importance of the $i$-th column of \texttt{x}; importance scale is arbitrary, but it must be a finite real number and rise with feature importance.
Adapters shipped with the package are, by convention, given names starting with \texttt{getImp}, with the default called \texttt{getImpRfZ}.

Transdapters are functions which wrap the adapter function with additional behaviours --- they may intercept and modify the transient data set produced for the adapter, as well as modify the returned importance scores.
They are implemented as functions expecting at least one argument, the adapter function, and immediately returning a function wrapping the given adapter.

One can write an example no-op transdapter in the following way:
<<noop>>=
library(Boruta)

noopTransdapter<-function(adapter=getImpRfZ){
 adapter
}
@

\ldots and use it this way:
<<noopuse>>=
set.seed(17)
Boruta(Species~.,iris,getImp=noopTransdapter())
@

\noindent As expected, this transdpater has not altered the Boruta result.

The Boruta package ships with several experimental transdapters.
Let us now briefly discuss each of them.

\section{Impute transdapter}
Many potential importance sources, in particular the default ranger \cite{Wright2015} implementation of Random Forest, can't handle missing values (\texttt{NA}s) in its input.
One can avoid this problem by performing imputation, that is a process of filling missing values with substitutes generated in some way, depending on the analysis aims and assumptions.
The problem here is that such a process may spuriously increase the importance of imputed features, bringing bias to the Boruta output, in particular when the imputation process uses the decision feature for its inference.

The impute transdapter alleviates this problem by performing a very simple, stochastic imputation on input, yet repeated every time importance is requested.
This way, imputation biases have a chance to cancel out, moreover, the effect of imputation is also reflected on shadows, allowing Boruta to better quantify its impact.

Let us test this on the \texttt{srx} data:
<<srximp>>=
set.seed(17)
data(srx)

srx_na<-srx
# Randomly punch 25 holes in the SRX data
holes<-25
holes<-cbind(
 sample(nrow(srx),holes,replace=TRUE),
 sample(ncol(srx),holes,replace=TRUE)
)
srx_na[holes]<-NA
# Use impute transdapter to mitigate them with internal imputation
Boruta(Y~.,srx_na,getImp=imputeTransdapter(getImpRfZ))
@

The precise method of imputation is to fill gaps with a random sample of other values in a feature, taken with replacement. 
All-missing features are replaced with a vector of zeros, which is non-important by design and should be given an importance of zero by a typical importance source.

One should note that this method is best suited for data sets with only a minor fraction of missing values, preferably randomly distributed.
Also it rather underestimates the importance of incomplete features.

\section{Decohere transdapter}
The main gain from using Boruta is that it can take into account multivariate interactions between groups features and the decision.
Decoherence transdapter can be used to validate whether multivariate inference is required to identify a certain feature as relevant, or just its marginal interaction with the decision suffices.

The decoherence transformation is defined for categorical decisions only, and is performed by randomly mixing the order of values in each feature, yet only within decision classes.
This way, feature-decision joint distribution is preserved, but feature-feature interactions are asymptotically knocked out.

Let us try this on the \texttt{srx} data:
<<srxdeco>>=
set.seed(17)
Boruta(Y~.,srx,getImp=decohereTransdapter())
@
\noindent We can see that most originally confirmed features are no longer relevant after decoherence transform is applied, save for \texttt{AnB} and \texttt{AoB} of a substantial marginal importance.

One should note that this approach is a heuristic and should be treated with reservation.
For instance, it may reduce the estimated shadow importance and henceforth spuriously elevate originally rejected features to a confirmed status.

\section{Conditional transdapter}

The conditional transdapter applies the wrapped importance source to each stratum of a given stratification vector, and averages importance scores with weights proportional to strata sizes.
This allows one to control for a certain variable or inject additional information.

Note that each strata should be large enough to be a proper sample for the information source used; stratification into too many very small groups may lead to spurious results.
Moreover, be aware that certain importance sources may be incompatible with the way this transdapter collects importance from strata; additional normalisation might be necessary.

\bibliography{vig}
\end{document}

\documentclass{article}
%\VignetteIndexEntry{Importance transdapters}
\usepackage[utf8]{inputenc}

\title{Importance transdapters}
\author{Miron B. Kursa}

\begin{document}
\bibliographystyle{plain}
\maketitle

\section{Overview}
Boruta is essentially a significance test for the importance of features generated by some importance source.
This source is usually the normalised permutational importance from a Random Forest \cite{Breiman2001} model, although other options are possible.
The interface between Boruta and importance source is called \emph{importance adapter}, which is an R function fed into \texttt{getImp} argument of the \texttt{Boruta} function.
They should expect three arguments, \texttt{x}, for predictors table with shadows, \texttt{y}, for the decision, and \texttt{...}, for arbitrary additional arguments relayed from the Boruta call.
For instance, \texttt{Boruta(Species~.,iris,threads=1)} is going to pass the \texttt{threads} argument to the adapter function, which in turn will append it to its call to \texttt{fru}.
The adapter output is expected to be a numerical vector of the same length as \texttt{ncol(x)}, with $i$-th element being the importance of the $i$-th column of \texttt{x}; importance scale is arbitrary, but it must be a finite real number and rise with feature importance.
Adapters shipped with the package are, by convention, given names starting with \texttt{getImp}, with the default called \texttt{getImpFruZ}.

Transdapters are functions which wrap the adapter function with additional behaviours --- they may intercept and modify the transient data set produced for the adapter, as well as modify the returned importance scores.
They are implemented as functions expecting at least one argument, the adapter function, and immediately returning a function wrapping the given adapter.

One can write an example no-op transdapter in the following way:
<<noop>>=
library(Boruta)

noopTransdapter<-function(adapter=getImpFruZ){
 adapter
}
@

\ldots and use it this way:
<<noopuse>>=
set.seed(17)
Boruta(Species~.,iris,getImp=noopTransdapter())
@

\noindent As expected, this transdpater has not altered the Boruta result.

The Boruta package ships with several transdapters.
Let us now briefly discuss each of them.

\section{Impute transdapter}
Many potential importance sources, in particular the default fru implementation of Random Forest, can't handle missing values (\texttt{NA}s) in its input.
One can avoid this problem by performing imputation, that is a process of filling missing values with substitutes generated in some way, depending on the analysis aims and assumptions.
The problem here is that such a process may spuriously increase the importance of imputed features, bringing bias to the Boruta output, in particular when the imputation process uses the decision feature for its inference.

The impute transdapter alleviates this problem by performing a very simple, stochastic imputation on input, yet repeated every time importance is requested.
This way, imputation biases have a chance to cancel out, moreover, the effect of imputation is also reflected on shadows, allowing Boruta to better quantify its impact.

Let us test this on the \texttt{srx} data:
<<srximp>>=
set.seed(17)
data(srx)

srx_na<-srx
# Randomly punch 25 holes in the SRX data
holes<-25
holes<-cbind(
 sample(nrow(srx),holes,replace=TRUE),
 sample(ncol(srx),holes,replace=TRUE)
)
srx_na[holes]<-NA
# Use impute transdapter to mitigate them with internal imputation
Boruta(Y~.,srx_na,getImp=imputeTransdapter(getImpFruZ))
@

The precise method of imputation is to fill gaps with a random sample of other values in a feature, taken with replacement. 
All-missing features are replaced with a vector of zeros, which is non-important by design and should be given an importance of zero by a typical importance source.

One should note that this method is best suited for data sets with only a minor fraction of missing values, preferably randomly distributed.
Also it rather underestimates the importance of incomplete features.

\section{Decohere transdapter}
The main gain from using Boruta is that it can take into account multivariate interactions between groups features and the decision.
Decoherence transdapter can be used to validate whether multivariate inference is required to identify a certain feature as relevant, or just its marginal interaction with the decision suffices.

The decoherence transformation is defined for categorical decisions only, and is performed by randomly mixing the order of values in each feature, yet only within decision classes.
This way, feature-decision joint distribution is preserved, but feature-feature interactions are asymptotically knocked out.

Let us try this on the \texttt{srx} data:
<<srxdeco>>=
set.seed(17)
Boruta(Y~.,srx,getImp=decohereTransdapter())
@
\noindent We can see that most originally confirmed features are no longer relevant after decoherence transform is applied, save for \texttt{AnB} and \texttt{AoB} of a substantial marginal importance.

One should note that this approach is a heuristic and should be treated with reservation.
For instance, it may reduce the estimated shadow importance and henceforth spuriously elevate originally rejected features to a confirmed status.

\section{Conditional transdapter}

The conditional transdapter applies the wrapped importance source to each stratum of a given stratification vector, and averages importance scores with weights proportional to strata sizes.
This allows one to control for a certain variable or inject additional information.

Note that each strata should be large enough to be a proper sample for the information source used; stratification into too many very small groups may lead to spurious results.
Moreover, be aware that certain importance sources may be incompatible with the way this transdapter collects importance from strata; additional normalisation might be necessary.

\bibliography{vig}
\end{document}

\documentclass{article}
%\VignetteIndexEntry{Boruta for those in a hurry}
\usepackage[utf8]{inputenc}

\title{Boruta for those in a hurry}
\author{Miron B. Kursa}

\begin{document}
\bibliographystyle{plain}
\maketitle

\setkeys{Gin}{width=\textwidth}

\section{Overview}
Boruta \cite{Kursa2010} is a feature selection method; that is, it expects a standard information system you'd fed to a classifier, and judges which of the features are important and which are not.
Let's try it with a sample dataset, say \texttt{iris}.
To make things interesting, we will add some nonsense features to see if they get filtered out; to this end, we randomly mix the order of elements in each of the original features, wiping out its interaction with the decision, \texttt{iris\$Species}.

<<setGeneration>>=
set.seed(17)
data(iris)
irisE<-cbind(
 setNames(
  data.frame(apply(iris[,-5],2,sample)),
  sprintf("Nonsense%d",1:4)
 ),
 iris
)
@

\noindent Now, time for Boruta:

<<Boruta>>=
library(Boruta)
Boruta(Species~.,data=irisE)->BorutaOnIrisE
BorutaOnIrisE
@

As one can see, the method had \textit{rejected} nonsense features and \textit{confirmed} (retained) the original ones, as it was to be expected.
What is important is that Boruta does a sharp classification of features rather than ordering, which is in contrast to many other feature selection methods.
The other substantial difference is that Boruta is an \textit{all relevant} method, hence aims to find all features connected with the decision --- most other methods are of a \textit{minimal optimal} class, consequently aims to provide a possibly compact set of features which carry enough information for a possibly optimal classification on the reduced set \cite{Nilsson2007}.
What does it mean in practice is that Boruta will include redundant features, that is ones which carry information already contained in other features.

As an example, let's add a feature which contains all the information in the decision in a most accessible form --- namely, a copy of the decision, and push it into Boruta.

<<BorutaReduendancy>>=
irisR<-cbind(
 irisE,
 SpoilerFeature=iris$Species
)
Boruta(Species~.,data=irisR)
@

We see that \texttt{SpoilerFeature} has not supplanted any of the original features, despite making them fully redundant.
One may wonder, however, how come anyone would need something which is clearly redundant?
There are basically three reasons behind this:
\begin{itemize}
 \item One may perform feature selection for an insight in which aspects of the phenomenon in question are important and which are not.
 In such cases subtle effects possess substantial explanatory value, even if they are masked by stronger interactions.
 \item In some sets, especially of a $p\gg n$ class, nonsense features may have spurious correlations with the decision, arising purely by chance.
 Such interactions may rival or even be stronger than actual mechanisms of the underlying phenomenon, making them apparently redundant.
 All relevant approaches won't magically help distinguish both, but will better preserve true patterns.
 \item Minimal optimal methods are generally cherry-picking features usable for classification, regardless if this usability is significant or not, which is an easy way to overfitting.
 Boruta is much more robust in this manner.
\end{itemize}

\section{Mechanism}

Under the hood, Boruta uses feature importance scores which are provided by certain machine learning methods; in particular Random Forest \cite{Breiman2001}, which happens to be used by default (using the \texttt{fru} package implementation).
Such scores only contribute to the ranking of features, though --- to separate relevant features, we need some reference of what is a distribution of importance of an irrelevant feature.
To this end, Boruta uses \textit{shadow features} or \textit{shadows}, which are copies of original features but with randomly mixed values, so that their distribution remains the same yet their importance is wiped out.


As importance scoring is often stochastic and can be degraded due to a presence of shadows, the Boruta selection is a process.
In each iteration, first shadows are generated, and such an extended dataset is fed to an importance provider.
Original features' importance is then compared with the highest importance of a shadow; and those which score higher are given a \textit{hit}.
Accumulated hit counts are finally assessed; features which significantly outperform best shadow are claimed confirmed, while those which significantly under-perform best shadow are claimed rejected and removed from the set for all subsequent iterations.
This loop is illustrated on Figure~\ref{fig:algo}.
\begin{figure}
\includegraphics[width=\textwidth]{algo.pdf}
\caption{\label{fig:algo} Illustration of the main loop of the Boruta algorithm.}
\end{figure}

The algorithm stops when all features have an established decision, or when a pre-set maximal number of iterations (100 by default) is exhausted.
In the latter case, the remaining features are claimed \textit{tentative}.

The process can be observed live with \texttt{doTrace} argument set to 1 (report after each decision), 2 (report after each iteration) or 3 (also report hits); importances in each iteration are also stored in the \texttt{ImpHistory} element of the Boruta object.
The graphical summary of a run can be obtained using \texttt{plot} and \texttt{plotImpHistory} on the Boruta result object, as shown on Figure~\ref{fig:plots} for the extended iris example.
First function uses boxplots to show the distribution of features' importance over Boruta run, using colours to mark final decision; it also draws boxplots for the importance of worst, average and best shadow in each iteration (in blue).
Second function visualises the same data, but as a function of the iteration number.
The summary of feature importance and hit counts can be extracted using the \texttt{attStats} convenience function.

\begin{figure}
<<BorutaPlots,fig=TRUE,echo=FALSE,results=hide,width=10,height=5>>=
par(mfrow=c(1,2))
plot(BorutaOnIrisE)
plotImpHistory(BorutaOnIrisE)
@
\caption{\label{fig:plots} The result of calling \texttt{plot} (left) and \texttt{plotImpHistory} (right) on the \texttt{BorutaOnIrisE} object.}
\end{figure}

<<attStats>>=
attStats(BorutaOnIrisE)
@

\section{Importance sources}

Building Random Forest multiple times on substantially enlarged dataset may easily become very time consuming, especially for larger sets for which using Boruta makes most sense.
It is also possible that RF importance is not best suited to catch the information content of features because of the dataset specifics.

Anyhow, Boruta allows you to switch importance source to an arbitrary function which gets an information system and returns a vector of numeric importance scores of all features.
The package already includes adapters for several importance scorers and their various configurations; all start with a \texttt{getImp} prefix.

You can pass arguments to the importance provider by providing it to the Boruta call; for instance, fru, the default importance provider, makes use of all available CPU threads, which won't always be the optimal choice.
Setting \texttt{threads} in the Boruta call will cause it to relay this argument to the \texttt{fru} function, and hence limit the training process parallelism.

Importance adapters may also be modified by transdapters (functions with names ending with \texttt{Transdapter}) to achieve some additional goals; consult the transdapter vignette or manual for more details.

\section{Caveats}
Few things worth noting before using Boruta in production:
\begin{itemize}
\item Boruta is a heuristic; there are no strict guarantees about its output.
Whenever possible, try to assess its results, especially in terms of selection stability as classification accuracy may be deceiving \cite{Kursa2014}.
\item For datasets with lots of features, the default configuration of the importance source is likely insufficient; in the particular case of Random Forest the number of trees is often not large enough to allow the importance scores to stabilise, which in turn often leads to false negatives and unstable results.
\item Boruta is a strictly serial algorithm, and spends most time waiting for the importance provider --- hence, tweaking this element brings the best chance to speed up the selection.
If speed is a concern, one should also avoid the formula interface and directly pass predictor and decision parts of the information system.
\item Elimination of tentative features becomes practically impossible if they turn out to have very similar importance distribution to the best shadow, and the presence of such does not make the overall Boruta result useless.
\item Importance history for bigger problems may take an impractically huge amount of memory; hence its collection can be turned off with \texttt{holdHistory} argument of the \texttt{Boruta} function.
This will disable some functionality, though, most notably plotting.
\item Treatment of missing values and non-standard decision forms (like survival problems) depends on the capacity of the information source.
\item The original Boruta paper describes the 1.0 version, and the algorithm has undergone substantial changes since then, namely the initial, warm-up rounds were removed, the multiple testing correction was introduced, finally the nomenclature has been clarified.
\end{itemize}

\bibliography{vig}
\end{document}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trandapters.R
\name{conditionalTransdapter}
\alias{conditionalTransdapter}
\title{Conditional transdapter}
\usage{
conditionalTransdapter(groups, adapter = getImpFruZ)
}
\arguments{
\item{groups}{groups.}

\item{adapter}{importance adapter to transform.}
}
\value{
transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
}
\description{
Applies downstream importance source on a given object strata and averages their outputs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance.R
\name{getImpFru}
\alias{getImpFru}
\alias{getImpFruZ}
\alias{getImFruRaw}
\alias{getImpFruRaw}
\title{Fru Random Forest importance adapters}
\usage{
getImpFruZ(
  x,
  y,
  ntree = 500,
  trees = ntree,
  num.threads = 0,
  threads = num.threads,
  ...
)

getImpFruRaw(
  x,
  y,
  ntree = 500,
  trees = ntree,
  num.threads = 0,
  threads = num.threads,
  ...
)
}
\arguments{
\item{x}{data frame of predictors including shadows.}

\item{y}{response vector.}

\item{ntree}{Number of trees in the forest; copied into \code{\link[fru]{fru}}'s native \code{trees}, put to retain transparent compatibility with randomForest.}

\item{trees}{Number of trees in the forest, as according to \code{\link[fru]{fru}}'s nomenclature. If not given, set to \code{ntree} value. If both are given, \code{trees} takes precedence.}

\item{num.threads}{Number of computing threads to use; copied into \code{\link[fru]{fru}}'s native \code{threads}, put to retain transparent compatibility with ranger.}

\item{threads}{Number of computing threads to use, as according to \code{\link[fru]{fru}}'s nomenclature.
If not given, set to \code{num.threads}.
If both are given, \code{threads} takes precedence.
Default value of zero means all available threads.}

\item{...}{parameters passed to the underlying \code{\link[fru]{fru}} call; they are relayed from \code{...} of \code{\link{Boruta}}.}
}
\description{
Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
\code{getImpFruZ} generates default, normalized permutation importance while \code{getImpFruRaw} raw permutation importance.
Fru does not support Gini index importance.
}
\note{
In prior versions of Boruta, other implementations of Random Forest were used as a default; prior to 5.0.0, randomForest package was used, and the default adapter used back then is available as \code{\link{getImpLegacyRf}}.
Prior to 10.0, ranger package was used with \code{\link{getImpRfZ}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trandapters.R
\name{imputeTransdapter}
\alias{imputeTransdapter}
\title{Impute transdapter}
\usage{
imputeTransdapter(adapter = getImpFruZ)
}
\arguments{
\item{adapter}{importance adapter to transform.}
}
\value{
transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
}
\description{
Wraps the importance adapter to accept NAs in input.
}
\note{
An all-NA feature will be converted to all zeroes, which should be ok as a totally non-informative value with most methods, but it is not universally correct.
Ideally, one should avoid having such features in input altogether.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{plot.Boruta}
\alias{plot.Boruta}
\title{Plot Boruta object}
\usage{
\method{plot}{Boruta}(
  x,
  colCode = c("green", "yellow", "red", "blue"),
  sort = TRUE,
  whichShadow = c(TRUE, TRUE, TRUE),
  col = NULL,
  xlab = "Attributes",
  ylab = "Importance",
  ...
)
}
\arguments{
\item{x}{an object of a class Boruta.}

\item{colCode}{a vector containing colour codes for attribute decisions, respectively Confirmed, Tentative, Rejected and shadow.}

\item{sort}{controls whether boxplots should be ordered, or left in original order.}

\item{whichShadow}{a logical vector controlling which shadows should be drawn; switches respectively max shadow, mean shadow and min shadow.}

\item{col}{standard \code{col} attribute. If given, suppresses effects of \code{colCode}.}

\item{xlab}{X axis label that will be passed to \code{\link{boxplot}}.}

\item{ylab}{Y axis label that will be passed to \code{\link{boxplot}}.}

\item{...}{additional graphical parameter that will be passed to \code{\link{boxplot}}.}
}
\value{
Invisible copy of \code{x}.
}
\description{
Default plot method for Boruta objects, showing boxplots of attribute importances over run.
}
\note{
If \code{col} is given and \code{sort} is \code{TRUE}, the \code{col} will be permuted, so that its order corresponds to attribute order in \code{ImpHistory}.

This function will throw an error when \code{x} lacks importance history, i.e., was made with \code{holdHistory} set to \code{FALSE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{attStats}
\alias{attStats}
\title{Extract attribute statistics}
\usage{
attStats(x)
}
\arguments{
\item{x}{an object of a class Boruta, from which attribute stats should be extracted.}
}
\value{
A data frame containing, for each attribute that was originally in information system, mean, median, maximal and minimal importance, number of hits normalised to number of importance source runs performed and the decision copied from \code{finalDecision}.
}
\description{
\code{attStats} shows a summary of a Boruta run in an attribute-centred way.
It produces a data frame containing some importance stats as well as the number of hits that attribute scored and the decision it was given.
}
\note{
When using a Boruta object generated by a \code{\link{TentativeRoughFix}}, the resulting data frame will consist a rough-fixed decision.

\code{x} has to be made with \code{holdHistory} set to \code{TRUE} for this code to run.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance.R
\name{getImpFerns}
\alias{getImpFerns}
\title{Random Ferns importance}
\usage{
getImpFerns(x, y, ...)
}
\arguments{
\item{x}{data frame of predictors including shadows.}

\item{y}{response vector.}

\item{...}{parameters passed to the underlying \code{\link[rFerns]{rFerns}} call; they are relayed from \code{...} of \code{\link{Boruta}}.}
}
\description{
This function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
}
\note{
Random Ferns importance calculation should be much faster than using Random Forest; however, one must first optimize the value of the \code{depth} parameter and
it is quite likely that the number of ferns in the ensemble required for the importance to converge will be higher than the number of trees in case of Random Forest.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{getConfirmedFormula}
\alias{getConfirmedFormula}
\alias{getNonRejectedFormula}
\title{Export Boruta result as a formula}
\usage{
getConfirmedFormula(x)

getNonRejectedFormula(x)
}
\arguments{
\item{x}{an object of a class Boruta, made using a formula interface.}
}
\value{
Formula, corresponding to the Boruta results.
\code{getConfirmedFormula} returns only Confirmed attributes, \code{getNonRejectedFormula} also adds Tentative ones.
}
\description{
Functions which convert the Boruta selection into a formula, so that it could be passed further to other functions.
}
\note{
This operation is possible only when Boruta selection was invoked using a formula interface.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance.R
\name{getImpRf}
\alias{getImpRf}
\alias{getImpRfZ}
\alias{getImpRfGini}
\alias{getImpRfRaw}
\title{Ranger Random Forest importance adapters}
\usage{
getImpRfZ(x, y, ntree = 500, num.trees = ntree, ...)

getImpRfGini(x, y, ntree = 500, num.trees = ntree, ...)

getImpRfRaw(x, y, ntree = 500, num.trees = ntree, ...)
}
\arguments{
\item{x}{data frame of predictors including shadows.}

\item{y}{response vector.}

\item{ntree}{Number of trees in the forest; copied into \code{\link[ranger]{ranger}}'s native num.trees, put to retain transparent compatibility with randomForest.}

\item{num.trees}{Number of trees in the forest, as according to \code{\link[ranger]{ranger}}'s nomenclature. If not given, set to \code{ntree} value. If both are given, \code{num.trees} takes precedence.}

\item{...}{parameters passed to the underlying \code{\link[ranger]{ranger}} call; they are relayed from \code{...} of \code{\link{Boruta}}.}
}
\description{
Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
\code{getImpRfZ} generates default, normalized permutation importance, \code{getImpRfRaw} raw permutation importance, finally \code{getImpRfGini} generates Gini index importance.
}
\note{
\code{getImpRfZ} was the default importance source between versions 5.0.0 and 10.0.0. Prior to Boruta 5.0, \code{getImpLegacyRfZ} function was a default; see \link{getImpLegacyRf} for more details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Boruta.R
\name{print.Boruta}
\alias{print.Boruta}
\title{Print Boruta object}
\usage{
\method{print}{Boruta}(x, ...)
}
\arguments{
\item{x}{an object of a class Boruta.}

\item{...}{additional arguments passed to \code{\link{print}}.}
}
\value{
Invisible copy of \code{x}.
}
\description{
Print method for the Boruta objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{getSelectedAttributes}
\alias{getSelectedAttributes}
\title{Extract names of the selected attributes}
\usage{
getSelectedAttributes(x, withTentative = FALSE)
}
\arguments{
\item{x}{an object of a class Boruta, from which relevant attributes names should be extracted.}

\item{withTentative}{if set to \code{TRUE}, Tentative attributes will be also returned.}
}
\value{
A character vector with names of the relevant attributes.
}
\description{
\code{getSelectedAttributes} returns a vector of names of attributes selected during a Boruta run.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/importance.R
\name{getImpLegacyRf}
\alias{getImpLegacyRf}
\alias{getImpLegacyRfZ}
\alias{getImpLegacyRfGini}
\alias{getLegacyImpRfRaw}
\alias{getImpLegacyRfRaw}
\title{randomForest importance adapters}
\usage{
getImpLegacyRfZ(x, y, ...)

getImpLegacyRfRaw(x, y, ...)

getImpLegacyRfGini(x, y, ...)
}
\arguments{
\item{x}{data frame of predictors including shadows.}

\item{y}{response vector.}

\item{...}{parameters passed to the underlying \code{\link[randomForest]{randomForest}} call; they are relayed from \code{...} of \code{\link{Boruta}}.}
}
\description{
Those function is intended to be given to a \code{getImp} argument of \code{\link{Boruta}} function to be called by the Boruta algorithm as an importance source.
\code{getImpLegacyRfZ} generates default, normalized permutation importance, \code{getImpLegacyRfRaw} raw permutation importance, finally \code{getImpLegacyRfGini} generates Gini index importance, all using \code{\link[randomForest]{randomForest}} as a Random Forest algorithm implementation.
}
\note{
The \code{getImpLegacyRfZ} function was a default importance source in Boruta versions prior to 5.0; since then \code{\link[ranger]{ranger}} Random Forest implementation is used instead of \code{\link[randomForest]{randomForest}}, for speed, memory conservation and an ability to use multithreading.
Both importance sources should generally lead to the same results, yet there are differences.

Most notably, ranger by default treats factor attributes as ordered (and works very slow if instructed otherwise with \code{respect.unordered.factors=TRUE}); on the other hand it lifts 32 levels limit specific to \code{\link[randomForest]{randomForest}}.
To this end, Boruta decision for factor attributes may be different.

Random Forest methods has two main parameters, number of attributes tried at each split and the number of trees in the forest; first one is called \code{mtry} in both implementations, but the second \code{ntree} in \code{\link[randomForest]{randomForest}} and \code{num.trees} in \code{\link[ranger]{ranger}}.
To this end, to maintain compatibility, \code{getImpRf*} functions still accept \code{ntree} parameter relaying it into \code{num.trees}.
Still, both parameters take the same defaults in both implementations (square root of the number all all attributes and 500 respectively).

Moreover, \code{\link[ranger]{ranger}} brings some addition capabilities to Boruta, like analysis of survival problems or sticky variables which are always considered on splits.

Finally, the results for the same PRNG seed will be different.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{TentativeRoughFix}
\alias{TentativeRoughFix}
\title{Rough fix of Tentative attributes}
\usage{
TentativeRoughFix(x, averageOver = Inf)
}
\arguments{
\item{x}{an object of a class Boruta.}

\item{averageOver}{Either number of last importance source runs to
average over or Inf for averaging over the whole Boruta run.}
}
\value{
A Boruta class object with modified \code{finalDecision} element.
Such object has few additional elements:
\item{originalDecision}{Original \code{finalDecision}.}
\item{averageOver}{Copy of \code{averageOver} parameter.}
}
\description{
In some circumstances (too short Boruta run, unfortunate mixing of shadow attributes, tricky dataset\ldots), Boruta can leave some attributes Tentative.
\code{TentativeRoughFix} performs a simplified, weaker test for judging such attributes.
}
\details{
Function claims as Confirmed those attributes that
have median importance higher than the median importance of
maximal shadow attribute, and the rest as Rejected.
Depending of the user choice, medians for the test
are count over last round, all rounds or N last
importance source runs.
}
\note{
This function should be used only when strict decision is
highly desired, because this test is much weaker than Boruta
and can lower the confidence of the final result.

\code{x} has to be made with \code{holdHistory} set to
\code{TRUE} for this code to run.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Boruta.R
\docType{data}
\name{srx}
\alias{srx}
\title{Small redundant XOR data}
\format{
A data frame with 8 predictors, 4 relevant: A, B, AoB, AnB and nA, as well as 3 irrelevant N1, N2 and N3, and decision attribute Y.
}
\source{
\url{https://blog.mbq.me/relevance-and-redundancy/}
}
\usage{
data(srx)
}
\description{
A synthetic data set with 32 rows corresponding to all combinations of values of five logical features, A, B, N1, N2 and N3.
The decision Y is equal to A xor B, hence N1--N3 are irrelevant attributes.
The set also contains 3 additional features, A or B (AoB), A and B (AnB) and not A (nA), which provide a redundant, but still relevant way to reconstruct Y.
}
\details{
This is set is an easy way to demonstrate the difference between all relevant feature selection methods, which should select all features except N1--N3, and minimal optimal ones, which will probably ignore most of them.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Boruta.R
\name{Boruta}
\alias{Boruta}
\alias{Boruta.default}
\alias{Boruta.formula}
\title{Feature selection with the Boruta algorithm}
\usage{
Boruta(x, ...)

\method{Boruta}{default}(
  x,
  y,
  pValue = 0.01,
  mcAdj = TRUE,
  maxRuns = 100,
  doTrace = 0,
  holdHistory = TRUE,
  getImp = getImpFruZ,
  ...
)

\method{Boruta}{formula}(formula, data, ...)
}
\arguments{
\item{x}{data frame of predictors.}

\item{...}{additional parameters passed to \code{getImp}.}

\item{y}{response vector; factor for classification, numeric vector for regression, \code{Surv} object for survival (support depends on importance adapter capabilities).}

\item{pValue}{confidence level. Default value should be used.}

\item{mcAdj}{if set to \code{TRUE}, a multiple comparisons adjustment using the Bonferroni method will be applied. Default value should be used; older (1.x and 2.x) versions of Boruta were effectively using \code{FALSE}.}

\item{maxRuns}{maximal number of importance source runs.
You may increase it to resolve attributes left Tentative.}

\item{doTrace}{verbosity level. 0 means no tracing, 1 means reporting decision about each attribute as soon as it is justified, 2 means the same as 1, plus reporting each importance source run, 3 means the same as 2, plus reporting of hits assigned to yet undecided attributes.}

\item{holdHistory}{if set to \code{TRUE}, the full history of importance is stored and returned as the \code{ImpHistory} element of the result.
Can be used to decrease a memory footprint of Boruta in case this side data is not used, especially when the number of attributes is huge; yet it disables plotting of such made \code{Boruta} objects and the use of the \code{\link{TentativeRoughFix}} function.}

\item{getImp}{function used to obtain attribute importance, called importance adapter in this documentation.
The default is getImpFruZ, which runs Random Forest implementation from the \code{fru} package and gathers Z-scores of mean decrease accuracy measure.
It should return a numeric vector of a size identical to the number of columns of its first argument, containing importance measure of respective attributes.
Any order-preserving transformation of this measure will yield the same result.
It is assumed that more important attributes get higher importance. +-Inf are accepted, NaNs and NAs are treated as 0s, with a warning.}

\item{formula}{alternatively, formula describing model to be analysed.}

\item{data}{in which to interpret formula.}
}
\value{
An object of class \code{Boruta}, which is a list with the following components:
\item{finalDecision}{a factor of three value: \code{Confirmed}, \code{Rejected} or \code{Tentative}, containing final result of feature selection.}
\item{ImpHistory}{a data frame of importances of attributes gathered in each importance source run.
Beside predictors' importances, it contains maximal, mean and minimal importance of shadow attributes in each run.
Rejected attributes get \code{-Inf} importance.
Set to \code{NULL} if \code{holdHistory} was given \code{FALSE}.}
\item{timeTaken}{time taken by the computation.}
\item{impSource}{string describing the source of importance, equal to a comment attribute of the \code{getImp} argument.}
\item{call}{the original call of the \code{Boruta} function.}
}
\description{
Boruta is an all relevant feature selection wrapper algorithm, capable of working with any classification method that output variable importance measure (VIM); by default, Boruta uses Random Forest.
The method performs a top-down search for relevant features by comparing original attributes' importance with importance achievable at random, estimated using their permuted copies, and progressively eliminating irrelevant features to stabilise that test.
}
\details{
Boruta iteratively compares importances of attributes with importances of shadow attributes, created by shuffling original ones.
Attributes that have significantly worst importance than shadow ones are being consecutively dropped.
On the other hand, attributes that are significantly better than shadows are admitted to be Confirmed.
Shadows are re-created in each iteration.
Algorithm stops when only Confirmed attributes are left, or when it reaches \code{maxRuns} importance source runs.
If the second scenario occurs, some attributes may be left without a decision.
They are claimed Tentative.
You may try to extend \code{maxRuns} or lower \code{pValue} to clarify them, but in some cases their importances do fluctuate too much for Boruta to converge.
Instead, you can use \code{\link{TentativeRoughFix}} function, which will perform other, weaker test to make a final decision, or simply treat them as undecided in further analysis.
}
\references{
Miron B. Kursa, Witold R. Rudnicki (2010). Feature Selection with the Boruta Package.
\emph{Journal of Statistical Software, 36(11)}, p. 1-13.
URL: \doi{10.18637/jss.v036.i11}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trandapters.R
\name{decohereTransdapter}
\alias{decohereTransdapter}
\title{Decohere transdapter}
\usage{
decohereTransdapter(adapter = getImpFruZ)
}
\arguments{
\item{adapter}{importance adapter to transform.}
}
\value{
transformed importance adapter which can be fed into \code{getImp} argument of the \code{\link{Boruta}} function.
}
\description{
Applies the decoherence transformation to the input, destroying all multivariate interactions.
It will trash the Boruta result, only apply if you know what are you doing!
Works only for categorical decision.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{plotImpHistory}
\alias{plotImpHistory}
\title{Plot Boruta object as importance history}
\usage{
plotImpHistory(
  x,
  colCode = c("green", "yellow", "red", "blue"),
  col = NULL,
  type = "l",
  lty = 1,
  pch = 0,
  xlab = "Classifier run",
  ylab = "Importance",
  ...
)
}
\arguments{
\item{x}{an object of a class Boruta.}

\item{colCode}{a vector containing colour codes for attribute decisions, respectively Confirmed, Tentative, Rejected and shadow.}

\item{col}{standard \code{col} attribute, passed to \code{\link{matplot}}. If given, suppresses effects of \code{colCode}.}

\item{type}{Plot type that will be passed to \code{\link{matplot}}.}

\item{lty}{Line type that will be passed to \code{\link{matplot}}.}

\item{pch}{Point mark type that will be passed to \code{\link{matplot}}.}

\item{xlab}{X axis label that will be passed to \code{\link{matplot}}.}

\item{ylab}{Y axis label that will be passed to \code{\link{matplot}}.}

\item{...}{additional graphical parameter that will be passed to \code{\link{matplot}}.}
}
\value{
Invisible copy of \code{x}.
}
\description{
Alternative plot method for Boruta objects, showing matplot of attribute importances over run.
}
\note{
This function will throw an error when \code{x} lacks importance history, i.e., was made with \code{holdHistory} set to \code{FALSE}.
}
